"""
Constants for seq2seq models.
"""

PAD = "<PAD>"
PAD_ID = 0
UNK = "<UNK>"
UNK_ID = 1
SOS = "<SOS>"
SOS_ID = 2
EOS = "<EOS>"
EOS_ID = 3

VOCAB_PREFIX = [PAD, UNK, SOS, EOS]

EMB_INIT_RANGE = 1.0
INFINITY_NUMBER = 1e12
